'''

Copyright (c) 2017-2018 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

The functions below assume a configuration file called deid, although the
user can specify a custom name.

'''

from deid.logger import bot
from deid.utils import read_file
from deid.data import data_base
from deid.config.standards import (
    formats,
    actions,
    sections,
    filters
)
import os
import re
import sys
from collections import OrderedDict


def load_combined_deid(deids):
    '''load one or more deids, either based on a path or a tag
    
       Parameters
       ==========
       deids: should be a custom list of deids

    '''
    if not isinstance(deids,list):
        bot.warning("load_combined_deids expects a list.")
        sys.exit(1)

    found_format = None
    deid = None

    for single_deid in deids:
  
        # If not a tag or path, returns None
        next_deid = get_deid(tag=single_deid,
                             exit_on_fail=False,
                             quiet=True,
                             load=True)

        if next_deid is not None:

            # Formats must match
            if found_format is None:
                found_format = next_deid['format']
            else:
                if found_format != next_deid['format']:
                    bot.error('Mismatch in deid formats, %s and %s' %(found_format,
                                                                      next_deid['format']))
                    sys.exit(1)

            # If it's the first one, use as starter template
            if deid is None:
                deid = next_deid
            else:

                # Update filter, appending to end to give first preference
                if "filter" in next_deid:
                    if "filter" not in deid:
                        deid['filter'] = next_deid['filter']
                    else:
                        for name,group in next_deid['filter'].items():
                            if name in deid['filter']:
                                deid['filter'][name] = deid['filter'][name] + group
                            else:
                                deid['filter'][name] = group                      

                if "header" in next_deid:
                    if "header" not in deid:
                        deid['header'] = next_deid['header']
                    else:
                        deid['header'] = deid['header'] + next_deid['header']

        else:
            bot.warning('Problem loading %s, skipping.' %single_deid)
    return deid


def load_deid(path=None):
    '''load_deid will return a loaded in (user) deid configuration file
       that can be used to update a default config.json. If a file path is
       specified, it is loaded directly. If a folder is specified, we look
       for a deid file in the folder. If nothing is specified, we assume
       the user wants to load a deid file in the present working directory.
       If the user wants to have multiple deid files in a directory, this
       can be done with an extension that specifies the module, eg;
   
             deid.dicom
             deid.nifti

       Parameters
       ==========
       path: a path to a deid file

       Returns
       =======
       config: a parsed deid (dictionary) with valid sections

    '''
    path = find_deid(path)

    # Read in spec, clean up extra spaces and newlines
    spec = [x.strip('\n').strip(' ') 
           for x in read_file(path) 
           if x.strip('\n').strip(' ') not in ['']]

    spec = [x for x in spec if x not in ['', None]]
    config = OrderedDict()
    section = None

    while len(spec) > 0:

        # Clean up white trailing/leading space
        line = spec.pop(0).strip()

        # Comment
        if line.startswith("#"):
            continue

        # Starts with Format?
        elif bool(re.match('format', line, re.I)):
            fmt = re.sub('FORMAT|(\s+)','',line).lower()
            if fmt not in formats:
                bot.error("%s is not a valid format." %fmt)
                sys.exit(1)
            # Set format
            config['format'] = fmt
            bot.debug("FORMAT set to %s" %fmt)

        # A new section?
        elif line.startswith('%'):

            # Remove any comments
            line = line.split('#',1)[0].strip()

            # Is there a section name?
            section_name = None
            parts = line.split(' ')
            if len(parts) > 1:
                section_name = ' '.join(parts[1:])          
            section = re.sub('[%]|(\s+)','',parts[0]).lower()
            if section not in sections:
                bot.error("%s is not a valid section." %section)
                sys.exit(1)
            config = add_section(config=config,
                                 section=section,
                                 section_name=section_name)

        # An action (replace, blank, remove, keep, jitter)
        elif line.upper().startswith(actions):

            # Start of a filter group
            if line.upper().startswith('LABEL') and section == "filter":
                members = []
                keep_going = True
                while keep_going is True:
                    next_line = spec[0]                
                    if next_line.upper().strip().startswith('LABEL'):
                        keep_going = False
                    elif next_line.upper().strip().startswith("%"):
                        keep_going = False
                    else:
                        new_member = spec.pop(0)
                        members.append(new_member)
                    if len(spec) == 0:
                        keep_going = False

                # Add the filter label to the config
                config = parse_label(config=config,
                                     section=section,
                                     label=line,
                                     section_name=section_name,
                                     members=members)
            # Parse the action
            else:
                config = parse_action(section=section,
                                      section_name=section_name,
                                      line=line,
                                      config=config)
        else:
            bot.debug("%s not recognized to be in valid format, skipping." %line)
    return config


def find_deid(path=None):
    '''find_deid is a helper function to load_deid to find a deid file in
       a folder, or return the path provided if it is the file.

       Parameters
       ==========
       path: a path on the filesystem. If not provided, will assume PWD.

    '''
    if path is None:
        path = os.getcwd()

    # The user has provided a directory
    if os.path.isdir(path):
        contenders = ["%s/%s" %(path,x) for x 
                      in os.listdir(path)
                      if x.startswith('deid')]

        if len(contenders) == 0:
            bot.error("No deid settings files found in %s, exiting." %(path))
            sys.exit(1)

        elif len(contenders) > 1:   
            bot.warning("Multiple deid files found in %s, will use first." %(path))

        path = contenders[0]


    # We have a file path at this point
    if not os.path.exists(path):
        bot.error("Cannot find deid file %s, exiting." %(path))
        sys.exit(1)

    return path


def parse_label(section,config,section_name,members,label=None):
    '''parse label will add a (optionally named) label to the filter
       section, including one or more criteria
   
       Parameters
       ==========
       section: the section name (e.g., header) must be one in sections
       config: the config (dictionary) parsed thus far
       section_name: an optional name for a section
       members: the lines beloning to the section/section_name
       label: an optional name for the group of commands
    '''
    criteria = {'filters':[],
                'coordinates':[]}

    if label is not None:                                   
        label = (label.replace('label','',1)
                      .split('#')[0]
                      .strip())
        criteria['name'] = label

    while len(members) > 0:
        member = members.pop(0).strip()

        # We have a coordinate line
        if member.lower().startswith('coordinates'):
            coordinate = member.lower().replace('coordinates','').strip()
            criteria['coordinates'].append(coordinate)
            continue
        operator = None
        entry = None
        if member.startswith('+'):
            operator = 'and'
            member = member.replace('+','',1).strip()
        elif member.startswith('||'):
            operator = 'or'
            member = member.replace('||','',1).strip()

        # Now that operators removed, parse member
        if not member.lower().startswith(filters):
            bot.warning('%s filter is not valid, skipping.' %member.lower())
        else:

            # Returns single member with field, values, operator,
            # Or if multiple or/and in statement, a list
            entry = parse_member(member, operator)
        if entry is not None:
            criteria['filters'].append(entry.copy())

    config[section][section_name].append(criteria)
    return config



def parse_member(members, operator=None):

    main_operator = operator

    actions = []
    values = []
    fields = []
    operators = []
    members = [members]

    while len(members) > 0:

        operator = None
        value = None
        member = members.pop(0).strip()

        # Find the first || or +
        match_or=re.search('\|\|',member)
        match_and=re.search('\+',member)

        if match_or is not None:
            operator = "||"
        if match_and is not None:
            if match_or is not None:
                if match_or.start() >= match_and.start():
                    operator = "+" 
            else:
                operator = "+"

        if operator is not None:

            member,rest = member.split(operator,1)

            # The rest is only valid if contains a filter statement
            if any(word in rest for word in filters):
                members.append(rest.strip())

                # Split the statement based on found operator
                operator = (operator.replace('||','or')
                                    .replace('+', 'and'))
                operators.append(operator)
            else:
                member = operator.join([member,rest])

        # Parse the member
        action, member = member.split(' ',1)
        action = action.lower().strip()

        # Contains, equals, not equals expects FieldName Values
        if action in ['contains','equals','notequals']:
            try:
                field,value = member.split(' ',1)
            except ValueError:
                bot.error('%s for line %s must have field and values, exiting.' %(action,member))
                sys.exit(1)

        # Missing, empty, notcontains expect only a field
        elif action in ['missing', 'empty','notcontains', 'present']:
            field = member.strip()
        else:
            bot.error('%s is not a valid filter action.' %action)
            sys.exit(1)

        actions.append(action)
        fields.append(field.strip())

        if value is not None:
            values.append(value.strip())  
        
    entry = {'action':actions,
             'field':fields,
             'operator': main_operator,
             'InnerOperators':operators,
             'value': values }
    return entry


def add_section(config,section,section_name=None):
    '''add section will add a section (and optionally)
       section name to a config

       Parameters
       ==========
       config: the config (dict) parsed thus far
       section: the section name to add
       section_name: an optional name, added as a level

    '''

    if section is None:
        bot.error('You must define a section (e.g. %header) before any action.')
        sys.exit(1)

    if section == 'filter' and section_name is None:
        bot.error("You must provide a name for a filter section.")
        sys.exit(1)

    if section not in sections:
        bot.error("%s is not a valid section." %section)
        sys.exit(1)
            
    if section not in config:

        # If a section is named, we have more one level (dict)
        if section_name is not None:
            config[section] = OrderedDict()
            config[section][section_name] = []
            bot.debug("Adding section %s %s" %(section, section_name))
        else:
            config[section] = []
            bot.debug("Adding section %s" %section)
        return config

    # Section is in config
    if section_name is not None and section_name not in config[section]:
        config[section][section_name] = []

    return config


def parse_action(section, line, config, section_name=None):
    '''add action will take a line from a deid config file, a config (dictionary), and
       an active section name (eg header) and add an entry to the config file to perform
       the action.

       Parameters
       =========
       section: a valid section name from the deid config file
       line: the line content to parse for the section/action
       config: the growing/current config dictionary
       section_name: optionally, a section name

    '''
            
    if not line.upper().startswith(actions):
        bot.error("%s is not a valid action line." %line)
        sys.exit(1)

    # We may have to deal with cases of spaces
    parts = line.split(' ')
    action = parts.pop(0).replace(' ','')

    # What field is the action for?
    if len(parts) < 1:
        bot.error("%s requires a FIELD value, but not found." %(action))        
        sys.exit(1)

    field = parts.pop(0)

    # Actions that require a value
    if action in [ "ADD", "REPLACE", "JITTER" ]:
        if len(parts) == 0:
            bot.error("%s requires a VALUE, but not found" %(action))        
            sys.exit(1)
        value = ' '.join(parts[0:])  # get remained of line
        value = value.split('#')[0]  # remove comments
        bot.debug("Adding %s" %line) #
        config[section].append({ "action":action,
                                 "field":field,
                                 "value":value })

    # Actions that don't require a value
    elif action in [ "BLANK", "KEEP", "REMOVE" ]:
        bot.debug("%s: adding %s" %(section,line))
        config[section].append({ "action":action,
                                 "field":field })

    return config


def get_deid(tag=None, exit_on_fail=True, quiet=False, load=False):
    '''get deid is intended to retrieve the full path of a deid file provided with
       the software, based on a tag. For example, under deid/data if a file is called
       "deid.dicom", the tag would be "dicom". 

       Parameters
       ==========
       tag: the text that comes after deid to indicate the tag of the file in deid/data
       exit_on_fail: if None is an acceptable return value, this should be set to False
                     (default is True).
       quiet: Default False. If None is acceptable, quiet can be set to True
       load: also load the deid, if resulting path (from path or tag) is not None

    '''
    # no tag/path means load default
    if tag is None:
        tag = 'dicom'

    # If it's already loaded 
    if isinstance(tag,dict):
        bot.debug('deid is already loaded.')
        return tag

    # If it's a path, get full path
    if os.path.exists(tag):
        deid = os.path.abspath(tag)
    else:
        deid = "%s/deid.%s" %(data_base,tag)

    if not os.path.exists(deid):
        if quiet is False:
            bot.error("Cannot find %s" %(deid))
        if exit_on_fail is True:
            sys.exit(1)
        else:
            return None

    if load is True:
        return load_deid(deid)

    return deid



'''
clean.py: functions for pixel scrubbing

Copyright (c) 2017-2018 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

from deid.config import DeidRecipe
from deid.logger import bot
from deid.utils import get_temporary_name
from pydicom import read_file
import matplotlib
matplotlib.use('pdf')
import os
import re

bot.level = 3

class DicomCleaner():
    '''take an input dicom file, check for burned pixels, and then clean,
       with option to save / output in multiple formats. This object should
       map to one dicom file, and the usage flow is the following:

       cleaner = DicomCleaner()
       summary = cleaner.detect(dicom_file)
      
       cleaner.clean()
    '''
    
    def __init__(self, output_folder=None,
                       add_padding=False,
                       margin=3,
                       deid=None,
                       font=None,
                       force=True):

        if output_folder is None:
            output_folder = get_temporary_name(prefix="clean")

        if font is None:
            font = self.default_font()
        self.font = font
        self.cmap = 'gray'
        self.output_folder = output_folder
        self.recipe = DeidRecipe(deid)
        self.results = None
        self.force = force

    def default_font(self):
        '''define the font style for saving png figures
           if a title is provided
        '''
        return {'family': 'serif',
                'color':  'darkred',
                'weight': 'normal',
                'size': 16}

    def detect(self, dicom_file):
        '''detect will initiate the cleaner for a new dicom file.
        '''
        from deid.dicom.pixels.detect import has_burned_pixels
        self.results = has_burned_pixels(dicom_file, 
                                         deid=self.recipe.deid,
                                         force=self.force)
        self.dicom_file = dicom_file
        return self.results
        

    def clean(self):
        '''
        take a dicom image and a list of pixel coordinates, and return
        a cleaned file (if output file is specified) or simply plot 
        the cleaned result (if no file is specified)
    
        Parameters
        ==========
            add_padding: add N=margin pixels of padding
            margin: pixels of padding to add, if add_padding True
        '''

        if not self.results:
            bot.warning('Use %s.detect() to find coordinates first.' %self)

        else:
            bot.info('Scrubbing %s.' %self.dicom_file)

            # Load in dicom file, and image data
            dicom = read_file(self.dicom_file, force=True)

            # We will set original image to image, cleaned to clean
            self.original = dicom._get_pixel_array()
            self.cleaned = self.original.copy()

            # Compile coordinates from result
            coordinates = []
            for item in self.results['results']:
                if len(item['coordinates']) > 0:
                    for coordinate_set in item['coordinates']:
                        # Coordinates expected to be list separated by commas
                        new_coordinates = [int(x) for x in coordinate_set.split(',')]
                        coordinates.append(new_coordinates) # [[1,2,3,4],...[1,2,3,4]]

            for coordinate in coordinates:
                minr, minc, maxr, maxc = coordinate
                self.cleaned[minc:maxc, minr:maxr] = 0  # should fill with black
                                           

    def get_figure(self, show=False, image_type="cleaned", title=None):
        '''get a figure for an original or cleaned image. If the image
           was already clean, it is simply a copy of the original.
           If show is True, plot the image.
        '''
        from matplotlib import pyplot as plt
        
        if hasattr(self, image_type):
            fig, ax = plt.subplots(figsize=(10, 6))      
            ax.imshow(self.cleaned, cmap=self.cmap)
            if title is not None:
                plt.title(title, fontdict=self.font)
            if show is True:
                plt.show()
            return plt


    def _get_clean_name(self, output_folder, extension='dcm'):
        '''return a full path to an output file, with custom folder and
           extension. If the output folder isn't yet created, make it.
 
           Parameters
           ==========
           output_folder: the output folder to create, will be created if doesn't
           exist.
           extension: the extension of the file to create a name for, should
           not start with "."
        '''
        if output_folder is None:
            output_folder = self.output_folder

        if not os.path.exists(output_folder):
            bot.debug('Creating output folder %s' % output_folder)
            os.mkdir(output_folder)

        basename = re.sub('[.]dicom|[.]dcm', '', os.path.basename(self.dicom_file))
        return "%s/cleaned-%s.%s" %(output_folder, basename, extension)

    def save_png(self, output_folder=None, image_type="cleaned", title=None):
        '''save an original or cleaned dicom as png to disk.
           Default image_format is "cleaned" and can be set 
           to "original." If the image was already clean (not 
           flagged) the cleaned image is just a copy of original
        '''
        from matplotlib import pyplot as plt

        if hasattr(self, image_type):
            png_file = self._get_clean_name(output_folder, 'png')
            plt = self.get_figure(image_type=image_type, title=title)
            plt.savefig(png_file)
            plt.close()
            return png_file
        else:
            bot.warning('use detect() --> clean() before saving is possible.')


    def save_dicom(self, output_folder=None, image_type="cleaned"):
        '''save a cleaned dicom to disk. We expose an option to save
           an original (change image_type to "original" to be consistent,
           although this is not incredibly useful given it would duplicate
           the original data.
        '''
        # Having clean also means has dicom image
        if hasattr(self, image_type):
            dicom_name = self._get_clean_name(output_folder)
            dicom = read_file(self.dicom_file,force=True)
            # If going from compressed, change TransferSyntax
            if dicom.file_meta.TransferSyntaxUID.is_compressed is True:
                dicom.decompress()
            dicom.PixelData = self.cleaned.tostring()
            dicom.save_as(dicom_name)
            return dicom_name
        else:
            bot.warning('use detect() --> clean() before saving is possible.')



'''

Copyright (c) 2017-2018 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

from deid.config.standards import (
    actions as valid_actions
)

from .fields import expand_field_expression

from deid.logger import bot
from pydicom import read_file
from pydicom._dicom_dict import DicomDictionary
from deid.utils import (
    recursive_find,
    get_timestamp,
    parse_value
)
from .tags import *
import tempfile
import os
import re
import sys


# Actions

def perform_action(dicom, action, item=None, fields=None, return_seen=False):
    '''perform action takes  

       Parameters
       ==========
       dicom: a loaded dicom file (pydicom read_file)
       item: a dictionary with keys as fields, values as values
       fields: if provided, a filtered list of fields for expand
       action: the action from the parsed deid to take
          "deid" (eg, PatientID) the header field to process
          "action" (eg, REPLACE) what to do with the field
          "value": if needed, the field from the response to replace with
    '''
    field = action.get('field')   # e.g: PatientID, endswith:ID
    value = action.get('value')   # "suid" or "var:field"
    action = action.get('action') # "REPLACE"

    # If there is an expander applied to field, we iterate over
    fields = expand_field_expression(field=field,
                                     dicom=dicom,
                                     contenders=fields)

    # Keep track of fields we have seen
    seen = []    
    for field in fields:
        seen.append(field)
        dicom = _perform_action(dicom=dicom,
                                field=field,
                                item=item,
                                action=action,
                                value=value)
    if return_seen:
        return dicom, seen
    return dicom


def _perform_action(dicom, field, action, value=None, item=None):
    '''_perform_action is the base function for performing an action.
       perform_action (above) typically is called using a loaded deid,
       and perform_addition is typically done via an addition in a config
       Both result in a call to this function. If an action fails or is not
       done, None is returned, and the calling function should handle this.
    '''
    if action not in valid_actions:
        bot.warning('''%s in not a valid choice [%s]. 
                       Defaulting to blanked.''' %(action,
                                               ".".join(valid_actions)))
        action = "BLANK"

    if field in dicom and action != "ADD":

        # Blank the value
        if action == "BLANK":
            dicom = blank_tag(dicom,field)

        # Code the value with something in the response
        elif action == "REPLACE":

            value = parse_value(item, value, field)
            if value is not None:
                # If we make it here, do the replacement
                dicom = update_tag(dicom,
                                   field=field,
                                   value=value)
            else:
                bot.warning("REPLACE %s unsuccessful" %field)

        # Code the value with something in the response
        elif action == "JITTER":
            value = parse_value(item, value, field)
            if value is not None:

                # Jitter the field by the supplied value
                dicom = jitter_timestamp(item=dicom,
                                         field=field,
                                         value=value)
            else:
                bot.warning("JITTER %s unsuccessful" %field)

        # elif "KEEP" --> Do nothing. Keep the original

        # Remove the field entirely
        elif action == "REMOVE":
            dicom = remove_tag(dicom, field)

    elif action == "ADD":
        value = parse_value(item, value, field)
        if value is not None:
            dicom = add_tag(dicom, field, value, quiet=True) 

    return dicom


# Timestamps

def jitter_timestamp(dicom, field, value):
    '''if present, jitter a timestamp in dicom
       field "field" by number of days specified by "value"
       The value can be positive or negative.
 
       Parameters
       ==========
       dicom: the pydicom Dataset
       field: the field with the timestamp
       value: number of days to jitter by. Jitter bug!

    '''
    if not isinstance(value, int):
        value = int(value)

    original = dicom.get(field,None)
    if original is not None:
        dicom[field] = original + value
    return dicom



'''

Copyright (c) 2017-2018 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

from deid.logger import bot
from pydicom import read_file
from pydicom.tag import tag_in_exception
from pydicom.sequence import Sequence
from pydicom._dicom_dict import (
    DicomDictionary,
    RepeatersDictionary
)
from pydicom.tag import Tag
import os
import re
import sys

################################################################################
# Functions for Finding / Getting Tags
################################################################################

def get_tag(field):
    '''get_tag will return a dictionary with tag indexed by field. For each entry,
       a dictionary lookup is included with VR,  
    
       Parameters
       ==========
       field: the keyword to get tag for, eg "PatientIdentityRemoved"

    '''
    found = [{key:value} for key,value in DicomDictionary.items() if value[4] == field]
    tags = dict()
    if len(found) > 0:

        # (VR, VM, Name, Retired, Keyword
        found = found[0] # shouldn't ever have length > 1
        tag = Tag(list(found)[0])
        VR, VM, longName, retired, keyword = found[tag]

        manifest = {"tag": tag,
                    "VR": VR,
                    "VM":VM,
                    "keyword":keyword,
                    "name":longName }

        tags[field] = manifest 
    return tags


def find_tag(term, VR=None, VM=None, retired=False):
    '''find_tag will search over tags in the DicomDictionary and return the tags found
       to match some term.
    '''
    searchin = DicomDictionary
    if retired:
        searchin = RepeatersDictionary

    found = [value for key,value 
             in searchin.items() 
             if re.search(term,value[4]) or re.search(term,value[2])]

    # Filter by VR, VM, name, these are exact
    if VR is not None: found = _filter_tags(found,0,VR)
    if VM is not None: found = _filter_tags(found,1,VM)
    return found


def _filter_tags(tags, idx, fields=None):
    '''filter tags is a helper function to take some list of tags in the format
       [ (VR, VM, longname, retired, keyword).. ]
       where each of the items above has some index, idx, and filter that index
       down to what is provided in fields.
    '''
    if not isinstance(fields,list):
        fields = [fields]
    return [x for x in tags if x[idx] in fields]



################################################################################
# Manipulating Tags in Data
################################################################################

def remove_sequences(dicom):
    for field in dicom.dir():
        if isinstance(dicom.get(field),Sequence):
            dicom = remove_tag(dicom,field)
    return dicom


def add_tag(dicom, field, value, quiet=False):
    '''add tag will add a tag only if it's in the (active) DicomDictionary

       Parameters
       ==========
       dicom: the pydicom.dataset Dataset (pydicom.read_file)
       field: the name of the field to add
       value: the value to set, if name is a valid tag

    '''
    if quiet is False:
        bot.debug("Attempting ADDITION of %s." %(field))
    dicom = change_tag(dicom,field,value)
 
    # dicom.data_element("PatientIdentityRemoved")
    # (0012, 0062) Patient Identity Removed            CS: 'Yes'

    return dicom


def change_tag(dicom, field, value):
    '''change tag is a general function that can be used by 
       update_tag or add_tag. The only difference is the print output,
       and determining to call the function based on different conditions

       Parameters
       ==========
       dicom: the pydicom.dataset Dataset (pydicom.read_file)
       field: the name of the field to add
       value: the value to set, if name is a valid tag

    '''
    tag = get_tag(field)

    if field in tag:
        dicom.add_new(tag[field]['tag'], tag[field]['VR'], value) 
    else:
        bot.error("%s is not a valid field to add. Skipping." %(field))

    return dicom


def update_tag(dicom, field, value):
    '''update tag will update a value in the header, if it exists
       if not, nothing is added. This check is the only difference
       between this function and change_tag. 
       If the user wants to add a value (that might not exist) 
       the function add_tag should be used

       Parameters
       ==========
       dicom: the pydicom.dataset Dataset (pydicom.read_file)
       field: the name of the field to update
       value: the value to set, if name is a valid tag

    '''
    if field in dicom:
        dicom = change_tag(dicom, field, value)
    return dicom


def blank_tag(dicom, field):
    '''blank tag calls update_tag with value set to an
       empty string. If the tag cannot be found, warns the user
       and doesn't touch (in case of imaging data, or not found)

       Parameters
       ==========
       dicom: the pydicom.dataset Dataset (pydicom.read_file)
       field: the name of the field to blank

    '''
    # We cannot blank VR types of US or SS
    element = dicom.data_element(field)
    if element is not None:
        if element.VR not in ['US','SS']:
            return update_tag(dicom, field,"")
        bot.warning('Cannot determine tag for %s, skipping blank.' %field)
    return dicom


def remove_tag(dicom,field):
    '''remove tag will remove a tag if it is present in the dataset

       Parameters
       ==========
       dicom: the pydicom.dataset Dataset (pydicom.read_file)
       field: the name of the field to remove

    '''
    if field in dicom:
        tag = dicom.data_element(field).tag
        del dicom[tag]
    return dicom



#########################################################################
# Private Tags
#########################################################################


def get_private(dicom):
    '''get private tags

       Parameters
       ==========
       dicom: the pydicom.dataset Dataset (pydicom.read_file)
    '''
    datasets = [dicom]
    private_tags = []
    while len(datasets) > 0:
        ds = datasets.pop(0)
        taglist = sorted(ds.keys())
        for tag in taglist:
            with tag_in_exception(tag):
                if tag in ds:
                    try:
                        data_element = ds[tag]
                        if data_element.tag.is_private:
                            bot.debug(data_element.name)
                            private_tags.append(data_element)
                            if tag in ds and data_element.VR == "SQ":
                                sequence = data_element.value
                                for dataset in sequence:
                                    datasets.append(dataset)                        
                    except IndexError:
                        bot.debug("tag %s key present without value" %tag)
                    except NotImplementedError:
                        bot.debug('tag %s is invalid, skipping' %tag)
    return private_tags


def has_private(dicom):
    '''has_private will return True if the header has private tags

       Parameters
       ==========
       dicom: the pydicom.dataset Dataset (pydicom.read_file)

    '''
    private_tags = len(get_private(dicom))
    print("Found %s private tags" %private_tags)
    if private_tags > 0:
        return True
    return False



'''

Copyright (c) 2017-2018 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

from pydicom.dataset import Dataset
from pydicom.sequence import Sequence
from deid.logger import bot
import os
import re

# These filters are based off the the CTP Dicom Filter
# http://mircwiki.rsna.org/index.php?title=The_CTP_DICOM_Filter
# We don't apply them to the tag, as in their examples:

# !ImageType.contains("SECONDARY")

# we apply them to the dataset, with the tag as an argument:
# dicom.contains("ImageType","SECONDARY")

def apply_filter(dicom, field, filter_name, value):
    '''essentially a switch statement to apply a filter to a dicom file.

       Parameters
       ==========
       dicom: the pydicom.dataset Dataset (pydicom.read_file)
       field: the name of the field to apply the filter to
       filer_name: the name of the filter to apply (e.g., contains)
       value: the value to set, if filter_name is valid

    '''
    filter_name = filter_name.lower().strip()

    if filter_name == "contains":
        return dicom.contains(field,value)

    if filter_name == "notcontains":
        return dicom.notContains(field,value)

    elif filter_name == "equals":
        return dicom.equals(field,value)

    elif filter_name == "missing":
        return dicom.missing(field)

    elif filter_name == "present":
        return not dicom.missing(field)

    elif filter_name == "empty":
        return dicom.empty(field)

    elif filter_name == "notequals":
        return dicom.notEquals(field,value)

    bot.warning("%s is not a valid filter name, returning False" %filter_name)
    return False



################################################################################
# Equals
################################################################################


def equalsBase(self, field, term, ignore_case=True, not_equals=False):
    '''base of equals, with variable for ignore case (default True)
    '''

    is_equal = False

    contenders = self.get(field)

    if not isinstance(contenders,list):
        contenders = [contenders]    

    # In this loop we can only switch to True
    for contender in contenders:
        if contender is not None:
            if ignore_case:
                if not isinstance(contender,Sequence):                
                    try:
                        contender = str(contender).lower().strip()
                        term = str(term).lower().strip()
                    except AttributeError:
                        pass # we are dealing with number

                if contender == term:
                    is_equal = True

    # If we want to know not_equals, reverse
    if not_equals is True:
        is_equal = not is_equal

    return is_equal


def equals(self, field, term):
    '''returns true if the value of the identifier exactly 
       equals the string argument; otherwise, it returns false.'''
    return self.equalsBase(field,term)


def notEquals(self, field, term):
    return self.equalsBase(field=field,
                           term=term,
                           not_equals=True)



Dataset.equalsBase = equalsBase
Dataset.equals = equals
Dataset.notEquals = notEquals

################################################################################
# Empty and Null
#
# missing: means the field is not present (None)
# empty: means the field is present and empty
################################################################################


def missing(self, field):
    '''missing returns True if the dicom is missing the field entirely
       This means that the entire field is None
    '''
    content = self.get(field)
    if content == None:
        return True
    return False

 
def empty(self, field):
    '''empty returns True if the value is found to be ""
    '''
    content = self.get(field)
    if content == "":
        return True
    return False


Dataset.empty = empty
Dataset.missing = missing


################################################################################
# Matches and Contains
# 
# contains: searches across entire field
# matches: looks for exact match
################################################################################


def compareBase(self, field, expression, func, ignore_case=True):
    '''compareBase takes either re.search (for contains) or
       re.match (for matches) and returns True if the given regular
       expression is contained or matched
    '''
    is_match = False

    contenders = self.get(field)

    if not isinstance(contenders,list):
        contenders = [contenders]    

    for contender in contenders:
        if contender is not None:

            if not isinstance(contender,Sequence):                
                if ignore_case:
                    try:
                        contender = str(contender).lower().strip()
                        expression = str(expression).lower().strip()
                    except AttributeError:
                        pass # we are dealing with number
                             # sequence, or other private tag

                if func(expression,contender):
                    is_match = True

    return is_match


def matches(self, field, expression):
    '''matches returns true if the value of the identifier matches 
       the regular expression specified in the string argument; 
       otherwise, it returns false.
    '''
    return self.compareBase(field=field,
                            expression=expression,
                            func=re.match)


def contains(self, field, expression):
    '''contains returns true if the value of the identifier 
       contains the the string argument anywhere within it; 
       otherwise, it returns false.
    '''
    return self.compareBase(field=field,
                            expression=expression,
                            func=re.search)

def notContains(self, field, expression):
    '''notContains returns true if the value of the identifier 
       does not contain the the string argument anywhere within it; 
    '''
    return not self.compareBase(field=field,
                                expression=expression,
                                func=re.search)


Dataset.compareBase = compareBase
Dataset.matches = matches
Dataset.contains = contains
Dataset.notContains = notContains

################################################################################
# Starts and Endswith
################################################################################


def startsWith(self, field, term):
    '''startsWith returns true if the value of the identifier 
       starts with the string argument; otherwise, it returns false.
    '''
    expression = "^%s" %term
    return self.compareBase(field=field,
                            expression=expression,
                            func=re.match)
    
def endsWith(self, field, term):
    '''endsWith returns true if the value of the identifier ends with 
       the string argument; otherwise, it returns false.
    '''
    expression = "%s$" %term
    return self.compareBase(field=field,
                            expression=expression,
                            func=re.match)
    
Dataset.startsWith = startsWith
Dataset.endsWith = endsWith



from .general import (
    get_installdir,
    get_temporary_name,
    write_file,
    write_json,
    read_file,
    read_json,
    recursive_find,
    to_int
)

from .actions import (
    get_timestamp,
    get_func,
    parse_value
)



'''

Copyright (c) 2017-2018 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

from deid.config.standards import (
    actions as valid_actions
)

from .fields import expand_field_expression

from deid.logger import bot
from pydicom import read_file
from pydicom._dicom_dict import DicomDictionary
from pydicom.tag import Tag
from .tags import *
from .validate import validate_dicoms
from deid.utils import recursive_find
import tempfile
import os
import re
import sys

################################################################################
# Functions for Dicom files
################################################################################

def get_files(contenders, check=True, pattern=None, force=False):
    '''get_files will take a list of single dicom files or directories,
       and return a generator that yields complete paths to all files
    
       Parameters
       ==========
       conteners: a list of files or directories (contenders!)
       check: boolean to indicate if we should validate dicoms (default True)
       pattern: A pattern to use with fnmatch. If None, * is used
       force: force reading of the files, if some headers invalid.
              Not recommended, as many non-dicom will come through

    '''
    if not isinstance(contenders,list):
        contenders = [contenders]

    for contender in contenders:
        if os.path.isdir(contender):
            dicom_files = recursive_find(contender, pattern=pattern)
        else:
            dicom_files = [contender]

        for dicom_file in dicom_files:
            if dicom_file is not None:
                if check:
                    validated_files = validate_dicoms(dicom_file, force=force)
                else:
                    validated_files = [dicom_file]

                for validated_file in validated_files:
                    bot.debug("Found contender file %s" %(validated_file))
                    yield validated_file


def save_dicom(dicom, dicom_file, output_folder=None, overwrite=False):
    '''save_dicom will save a dicom file to an output folder,
       making sure to not overwrite unless the user has enforced it

       Parameters
       ==========
       dicom: the pydicon Dataset to save
       dicom_file: the path to the dicom file to save (we only use basename)
       output_folder: the folder to save the file to
       overwrite: overwrite any existing file? (default is False)

    '''

    if output_folder is None:
        if overwrite is False:
            output_folder = tempfile.mkdtemp()
        else:
            output_folder = os.path.dirname(dicom_file)

    dicom_name = os.path.basename(dicom_file)
    output_dicom = os.path.join(output_folder,dicom_name)
    dowrite = True
    if overwrite is False:
        if os.path.exists(output_dicom):
            bot.error("%s already exists, overwrite set to False. Not writing." %dicom_name)
            dowrite = False

    if dowrite:
        dicom.save_as(output_dicom)
    return output_dicom



'''

Copyright (c) 2017-2018 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

import fnmatch
import json
import os
import re
import requests
import tempfile
from deid.logger import bot
from collections import OrderedDict
import sys


# Python less than version 3 must import OSError
if sys.version_info[0] < 3:
    from exceptions import OSError

################################################################################
# Local commands and requests
################################################################################

def get_installdir():
    '''get_installdir returns the installation directory of the application
    '''
    return os.path.abspath(os.path.dirname(os.path.dirname(__file__)))


def get_temporary_name(prefix=None, ext=None):
    '''get a temporary name, can be used for a directory or file. This does so
       without creating the file, and adds an optional prefix
  
       Parameters
       ==========
       prefix: if defined, add the prefix after deid
       ext: if defined, return the file extension appended. Do not specify "."
    '''
    deid_prefix = 'deid-'
    if prefix:
        deid_prefix = 'deid-%s-' % prefix

    tmpname = os.path.join(tempfile.gettempdir(), 
                           '%s%s' % (deid_prefix,
                                     next(tempfile._get_candidate_names())))
    if ext:
        tmpname = '%s.%s' % (tmpname, ext)
    return tmpname


################################################################################
## FILE OPERATIONS #############################################################
################################################################################

def write_file(filename, content, mode="w"):
    '''write_file will open a file, "filename" and write content, "content"
       and properly close the file

       Parameters
       ==========
       filename: the name of the file to write to
       content: the content to write to file
       mode: the mode to open the file, defaults to write (w)

    '''
    with open(filename,mode) as filey:
        filey.writelines(content)
    return filename


def write_json(json_obj, filename, mode="w", print_pretty=True):
    '''write_json will (optionally,pretty print) a json object to file

       Parameters
       ==========
       json_obj: the dict to print to json
       filename: the output file to write to
       pretty_print: if True, will use nicer formatting   

    '''
    with open(filename, mode) as filey:
        if print_pretty == True:
            filey.writelines(json.dumps(json_obj, indent=4, separators=(',', ': ')))
        else:
            filey.writelines(json.dumps(json_obj))
    return filename



def read_file(filename, mode="r"):
    '''write_file will open a file, "filename" and write content, "content"
       and properly close the file

       Parameters
       ==========
       filename: the name of the file to write to
       mode: the mode to open the file, defaults to read (r)

    '''
    with open(filename, mode) as filey:
        content = filey.readlines()
    return content



def read_json(filename, mode="r", ordered_dict=False):
    '''read_json will open a file, "filename" and read the string as json

       Parameters
       ==========
       filename: the name of the file to write to
       mode: the mode to open the file, defaults to read (r)
       ordered_dict: If true, return an OrderedDict (default is False)

    '''
    with open(filename,mode) as filey:
        if ordered_dict is False:
            content = json.loads(filey.read())
        else:
            content = json.loads(filey.read(), object_pairs_hook=OrderedDict)
    return content


def recursive_find(base, pattern=None):
    '''recursive find will yield dicom files in all directory levels
       below a base path. It uses get_dcm_files to find the files in the bases.

       Parameters
       ==========
       base: the base directory to search
       pattern: a pattern to match. If None, defaults to "*"

    '''
    if pattern is None:
        pattern = "*"

    for root, dirnames, filenames in os.walk(base):
        for filename in fnmatch.filter(filenames, pattern):
            yield os.path.join(root, filename)


################################################################################
## DATA FORMATS ################################################################
################################################################################


def to_int(value):
    if not isinstance(value,int):
        value = int(float(value))
    return value


def is_number(value):
    '''is_number determines if the value for a field is numeric
    '''
    if isinstance(value,int):
        return True
    if isinstance(value,float):
        return True
    return False



'''

logger/message.py: Python logger base

Copyright (c) 2016-2018 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import os
import sys
from .spinner import Spinner

ABORT = -5
FLAG = -4
ERROR = -3
WARNING = -2
LOG = -1
INFO = 1
CUSTOM = 1
QUIET = 0
VERBOSE = VERBOSE1 = 2
VERBOSE2 = 3
VERBOSE3 = 4
DEBUG = 5

PURPLE = "\033[95m"
YELLOW = "\033[93m"
RED = "\033[91m"
DARKRED = "\033[31m"
CYAN = "\033[36m"

class DeidMessage:

    def __init__(self, MESSAGELEVEL=None):
        self.level = get_logging_level()
        self.history = []
        self.errorStream = sys.stderr
        self.outputStream = sys.stdout
        self.colorize = self.useColor()
        self.colors = {ABORT: DARKRED,
                       FLAG: RED,
                       ERROR: RED,    
                       WARNING: YELLOW,  
                       LOG: PURPLE,      
                       CUSTOM: PURPLE,       
                       DEBUG: CYAN,      
                       'OFF': "\033[0m", # end sequence
                       'CYAN':CYAN,
                       'PURPLE':PURPLE,
                       'RED':RED,
                       'DARKRED':DARKRED,
                       'YELLOW':YELLOW}
                       

    # Colors --------------------------------------------

    def useColor(self):
        '''useColor will determine if color should be added
        to a print. Will check if being run in a terminal, and
        if has support for asci'''
        COLORIZE = get_user_color_preference()
        if COLORIZE is not None:
            return COLORIZE
        streams = [self.errorStream, self.outputStream]
        for stream in streams:
            if not hasattr(stream, 'isatty'):
                return False
            if not stream.isatty():
                return False
        return True

    def addColor(self, level, text):
        '''addColor to the prompt (usually prefix) if terminal
        supports, and specified to do so'''
        if self.colorize:
            if level in self.colors:
                text = "%s%s%s" % (self.colors[level],
                                   text,
                                   self.colors["OFF"])
        return text

    def emitError(self, level):
        '''determine if a level should print to
        stderr, includes all levels but INFO and QUIET'''
        if level in [ABORT,
                     ERROR,
                     WARNING,
                     VERBOSE,
                     VERBOSE1,
                     VERBOSE2,
                     VERBOSE3,
                     DEBUG]:
            return True
        return False

    def emitOutput(self, level):
        '''determine if a level should print to stdout
        only includes INFO'''
        if level in [LOG,
                     INFO,
                     CUSTOM]:
            return True
        return False

    def isEnabledFor(self, messageLevel):
        '''check if a messageLevel is enabled to emit a level
        '''
        if messageLevel <= self.level:
            return True
        return False

    def emit(self, level, message, prefix=None, color=None):
        '''emit is the main function to print the message
        optionally with a prefix
        :param level: the level of the message
        :param message: the message to print
        :param prefix: a prefix for the message
        '''
        if color is None:
            color = level

        if prefix is not None:
            prefix = self.addColor(color, "%s " % (prefix))
        else:
            prefix = ""
            message = self.addColor(color, message)

        # Add the prefix
        message = "%s%s" % (prefix, message)

        if not message.endswith('\n'):
            message = "%s\n" % message

        # If the level is quiet, only print to error
        if self.level == QUIET:
            pass

        # Otherwise if in range print to stdout and stderr
        elif self.isEnabledFor(level):
            if self.emitError(level):
                self.write(self.errorStream, message)
            else:
                self.write(self.outputStream, message)

        # Add all log messages to history
        self.history.append(message)

    def write(self, stream, message):
        '''write will write a message to a stream,
        first checking the encoding
        '''
        if isinstance(message, bytes):
            message = message.decode('utf-8')
        stream.write(message)

    def get_logs(self, join_newline=True):
        ''''get_logs will return the complete history, joined by newline
        (default) or as is.
        '''
        if join_newline:
            return '\n'.join(self.history)
        return self.history


    def show_progress(
            self,
            iteration,
            total,
            length=40,
            min_level=0,
            prefix=None,
            carriage_return=True,
            suffix=None,
            symbol=None):
        '''create a terminal progress bar, default bar shows for verbose+
        :param iteration: current iteration (Int)
        :param total: total iterations (Int)
        :param length: character length of bar (Int)
        '''
        percent = 100 * (iteration / float(total))
        progress = int(length * iteration // total)

        if suffix is None:
            suffix = ''

        if prefix is None:
            prefix = 'Progress'

        # Download sizes can be imperfect, setting carriage_return to False
        # and writing newline with caller cleans up the UI
        if percent >= 100:
            percent = 100
            progress = length

        if symbol is None:
            symbol = "="

        if progress < length:
            bar = symbol * progress + '|' + '-' * (length - progress - 1)
        else:
            bar = symbol * progress + '-' * (length - progress)

        # Only show progress bar for level > min_level
        if self.level > min_level:
            percent = "%5s" % ("{0:.1f}").format(percent)
            output = '\r' + prefix + \
                " |%s| %s%s %s" % (bar, percent, '%', suffix)
            sys.stdout.write(output),
            if iteration == total and carriage_return:
                sys.stdout.write('\n')
            sys.stdout.flush()

    # Logging ------------------------------------------


    def abort(self, message):
        self.emit(ABORT, message, 'ABORT')

    def flag(self, message):
        self.emit(FLAG, message, 'FLAGGED')

    def error(self, message):
        self.emit(ERROR, message, 'ERROR')

    def warning(self, message):
        self.emit(WARNING, message, 'WARNING')

    def log(self, message):
        self.emit(LOG, message, 'LOG')

    def custom(self, prefix, message, color=PURPLE):
        self.emit(CUSTOM, message, prefix, color)

    def info(self, message):
        self.emit(INFO, message)

    def newline(self):
        return self.info("")

    def verbose(self, message):
        self.emit(VERBOSE, message, "VERBOSE")

    def verbose1(self, message):
        self.emit(VERBOSE, message, "VERBOSE1")

    def verbose2(self, message):
        self.emit(VERBOSE2, message, 'VERBOSE2')

    def verbose3(self, message):
        self.emit(VERBOSE3, message, 'VERBOSE3')

    def debug(self, message):
        self.emit(DEBUG, message, 'DEBUG')

    def is_quiet(self):
        '''is_quiet returns true if the level is under 1
        '''
        if self.level < 1:
            return False
        return True


    # Terminal ------------------------------------------

    def table(self, rows, col_width=2):
        '''table will print a table of entries. If the rows is 
        a dictionary, the keys are interpreted as column names. if
        not, a numbered list is used.
        '''

        labels = [str(x) for x in range(1,len(rows)+1)]
        if isinstance(rows, dict):
            labels = list(rows.keys())
            rows = list(rows.values())

        for row in rows: 
            label = labels.pop(0)
            label = label.ljust(col_width)
            message = "\t".join(row)
            self.custom(prefix=label,
                        message=message)
        


def get_logging_level():
    '''get_logging_level will configure a logging to standard out based on the user's
    selected level, which should be in an environment variable called
    MESSAGELEVEL. if MESSAGELEVEL is not set, the maximum level
    (5) is assumed (all messages).
    '''
    try:
        level = int(os.environ.get("MESSAGELEVEL", DEBUG))

    except ValueError:

        level = os.environ.get("MESSAGELEVEL", DEBUG)
        if level == "CRITICAL":
            return FLAG
        elif level == "ABORT":
            return ABORT
        elif level == "ERROR":
            return ERROR
        elif level == "WARNING":
            return WARNING
        elif level == "LOG":
            return LOG
        elif level == "INFO":
            return INFO
        elif level == "QUIET":
            return QUIET
        elif level.startswith("VERBOSE"):
            return VERBOSE3
        elif level == "LOG":
            return LOG
        elif level == "DEBUG":
            return DEBUG

    return level


def get_user_color_preference():
    COLORIZE = os.environ.get('DEID_COLORIZE', None)
    if COLORIZE is not None:
        COLORIZE = convert2boolean(COLORIZE)
    return COLORIZE


def convert2boolean(arg):
    '''convert2boolean is used for environmental variables that must be
    returned as boolean'''
    if not isinstance(arg, bool):
        return arg.lower() in ("yes", "true", "t", "1", "y")
    return arg


DeidMessage.spinner = Spinner()
bot = DeidMessage()



'''

detect.py: functions for pixel scrubbing

Copyright (c) 2017 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

from deid.config import DeidRecipe
from deid.logger import bot
from pydicom import read_file
from deid.dicom.filter import apply_filter

import os
import sys

def has_burned_pixels(dicom_files, force=True, deid=None):
    ''' has burned pixels is an entrypoint for has_burned_pixels_multi (for 
    multiple images) or has_burned_pixels_single (for one detailed repor)
    We will use the MIRCTP criteria (see ref folder with the
    original scripts used by CTP) to determine if an image is likely to have 
    PHI, based on fields in the header alone. This script does NOT perform
    pixel cleaning, but returns a dictionary of results (for multi) or one
    detailed result (for single)
    '''
    # if the user has provided a custom deid, load it
    if not isinstance(deid, DeidRecipe):
        if deid is None:
            deid = 'dicom'
        deid = DeidRecipe(deid)

    if isinstance(dicom_files,list):
        return _has_burned_pixels_multi(dicom_files, force, deid)
    return _has_burned_pixels_single(dicom_files, force, deid)



def _has_burned_pixels_multi(dicom_files, force, deid):
    '''return a summary dictionary with lists of clean, and then lookups
       for flagged images with reasons. The deid should be a deid recipe
       instantiated from deid.config.DeidRecipe. This function should not
       be called directly, but should be called from has_burned_pixels
    '''

    # Store decisions in lookup based on filter groups
    decision = {'clean':[],
                'flagged':{}}

    for dicom_file in dicom_files:
        result = _has_burned_pixels_single(dicom_file=dicom_file,
                                           force=force,
                                           deid=deid)


        if result['flagged'] is False:
            # In this case, group is None
            decision['clean'].append(dicom_file)
        else:
            decision['flagged'][dicom_file] = result

    return decision


def _has_burned_pixels_single(dicom_file, force, deid):

    '''has burned pixels single will evaluate one dicom file for burned in
    pixels based on 'filter' criteria in a deid. If deid is not provided,
    will use application default. The method proceeds as follows:

    1. deid is loaded, with criteria groups ordered from specific --> general
    2. image is run down the criteria, stops when hits and reports FLAG
    3. passing through the entire list gives status of pass
    
    The default deid has a greylist, whitelist, then blacklist

    Parameters
    =========
    dicom_file: the fullpath to the file to evaluate
    force: force reading of a potentially erroneous file
    deid: the full path to a deid specification. if not defined, only default used

    deid['filter']['dangerouscookie'] <-- filter list "dangerouscookie"

    --> This is what an item in the criteria looks like
        [{'coordinates': ['0,0,512,110'],
          'filters': [{'InnerOperators': [],
          'action': ['notequals'],
          'field': ['OperatorsName'],
          'operator': 'and',
          'value': ['bold bread']}],
        'name': 'criteria for dangerous cookie'}]

    
    Returns
    =======
    --> This is what a clean image looks like:
        {'flagged': False, 'results': []}


    --> This is what a flagged image looks like:
       {'flagged': True,
        'results': [
                      {'reason': ' ImageType missing  or ImageType empty ',
                       'group': 'blacklist',
                       'coordinates': []}
                   ]
        }
    '''

    dicom = read_file(dicom_file,force=force)
    dicom_name = os.path.basename(dicom_file)
        
    # Load criteria (actions) for flagging
    filters = deid.get_filters()
    if not filters:
        bot.error('Deid provided does not have %filter, exiting.')
        sys.exit(1)

    # Return list with lookup as dicom_file
    results = []
    global_flagged = False

    for name,items in filters.items():
        for item in items:
            flags = []

            descriptions = [] # description for each group across items

            for group in item['filters']:
                group_flags = []         # evaluation for a single line
                group_descriptions = []

                # You cannot pop from the list
                for a in range(len(group['action'])):
                    action = group['action'][a]
                    field = group['field'][a]
                    value = ''
                    if len(group['value']) > a:
                        value = group['value'][a]
                    flag = apply_filter(dicom=dicom,
                                        field=field,
                                        filter_name=action,
                                        value=value or None)
                    group_flags.append(flag)
                    description = "%s %s %s" %(field,action,value)
                    if len(group['InnerOperators']) > a:
                        inner_operator = group['InnerOperators'][a]
                        group_flags.append(inner_operator)
                        description = "%s %s" %(description,inner_operator)
                    group_descriptions.append(description)

                # At the end of a group, evaluate the inner group   
                flag = evaluate_group(group_flags)

                # "Operator" is relevant for the outcome of the list of actions 
                operator = ''
                if 'operator' in group:
                    if group['operator'] is not None:
                        operator = group['operator']
                        flags.append(operator)

                flags.append(flag)
                reason = ('%s %s' %(operator,' '.join(group_descriptions))).replace('\n',' ')
                descriptions.append(reason)

            group_name = ''
            if "name" in item:
                group_name = item['name']

            # When we parse through a group, we evaluate based on all flags
            flagged = evaluate_group(flags=flags)

            if flagged is True:
                global_flagged = True
                reason = ' '.join(descriptions)

                result = {'reason': reason,
                          'group': name,
                          'coordinates': item['coordinates'] }

                results.append(result)

    results = {'flagged': global_flagged,
               'results': results }
    return results


def evaluate_group(flags):
    '''evaluate group will take a list of flags (eg:

        [True, and, False, or, True]

    And read through the logic to determine if the image result
    is to be flagged. This is how we combine a set of criteria in
    a group to come to a final decision.
    '''
    flagged = False
    first_entry = True

    while len(flags) > 0:
        flag = flags.pop(0)
        if flag == "and":
            flag = flags.pop(0)
            flagged = flag and flagged
        elif flag == "or":
            flag = flags.pop(0)
            flagged = flag or flagged
        else:
            # If it's the first entry
            if first_entry is True:
                flagged = flag
            else:
                flagged = flagged and flag
        first_entry = False

    return flagged



'''

DeidRecipe

Copyright (c) 2017-2018 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

The functions below assume a configuration file called deid, although the
user can specify a custom name.


'''

from deid.config.utils import (
    load_deid,
    get_deid,
    load_combined_deid
)
from deid.config.standards import (
    actions,
    sections,
    formats
)

from deid.logger import bot
import os
import re

bot.level = 3

class DeidRecipe:
    '''Create and work with a deid recipe to filter and perform operations on
       a dicom header. Usage typically looks like:

       deid = 'dicom.deid'
       recipe = DeidRecipe(deid)
       
       If deid is None, the default provided by the application is used.

       Parameters
       ==========
       deid: the deid recipe (or recipes) files to use. If more than one
             is provided, should be done in order of preference for load
             (later in the list overrides earlier loaded).
       base: if True, load a default base (default_base) before custom
       default_base: the default base to load if "base" is True

    '''
    
    def __init__(self, deid=None, base=False, default_base='dicom'):

        # If deid is None, use the default
        if deid is None:
            bot.warning('No specification, loading default base deid.%s' %default_base)
            base = True

        self._init_deid(deid, base=base, default_base=default_base)

    def __str__(self):
        return '[deid]'

    def __repr__(self):
        return '[deid]'

    def load(self, deid):
        '''load a deid recipe into the object. If a deid configuration is
           already defined, append to that. 
        '''
        deid = get_deid(deid)
        if deid is not None:
 
            # Update our list of files
            self._files.append(deid)
            self.files = list(set(self.files))

            # Priority here goes to additional deid
            self.deid = load_combined_deid([self.deid, deid])

    def _get_section(self, name):
        '''return a section (key) in the loaded deid, if it exists
        '''
        section = None
        if self.deid is not None:
            if name in self.deid:
                section = self.deid[name]
        return section


    def get_format(self):
        '''return the format of the loaded deid, if one exists
        '''
        return self._get_section('format')


    def get_filters(self, name=None):
        '''return all filters for a deid recipe, or a set based on a name
        '''
        filters = self._get_section('filter')
        if name is not None and filters is not None:
            filters = filters[name]        
        return filters


    def ls_filters(self):
        '''list names of filter groups
        '''
        filters = self._get_section('filter')
        return list(filters.keys())

    def get_actions(self, action=None, field=None):
        '''get deid actions to perform on a header, or a subset based on a type

           A header action is a list with the following:
           {'action': 'REMOVE', 'field': 'AssignedLocation'},
 
           Parameters
           ==========
           action: if not None, filter to action specified
           field: if not None, filter to field specified

        '''
        header = self._get_section('header')
        if header is not None:
            if action is not None:
                action = action.upper()
                header = [x for x in header if x['action'].upper() == action]      
            if field is not None:
                field = field.upper()
                header = [x for x in header if x['field'].upper() == field]  

        return header


    def _init_deid(self, deid=None, base=False, default_base='dicom'):
        '''initalize the recipe with one or more deids, optionally including 
           the default. This function is called at init time. If you need to add
           or work with already loaded configurations, use add/remove 
    
           Parameters
           ==========
           deid: the deid recipe (or recipes) files to use. If more than one
                 is provided, should be done in order of preference for load
                 (later in the list overrides earlier loaded).
           default_base: load the default base before the user customizations. 

        '''
        if deid is None:
            deid = []

        if not isinstance(deid,list):
            deid = [deid]

        if base is True:
            deid.append(default_base)

        self._files = deid

        if len(deid) == 0:
            bot.info('You can add custom deid files with .load().')
        self.deid = load_combined_deid(deid)



#!/usr/bin/env python3

'''

The MIT License (MIT)

Copyright (c) 2017-2018 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

from glob import glob
from deid.version import __version__
import tempfile
import argparse
import sys
import os


def get_parser():
    parser = argparse.ArgumentParser(
    description="Deid (de-identification, anonymization) command line tool.")

    # Global Variables
    parser.add_argument('--quiet',"-q", dest="quiet", 
                        help="Quiet the verbose output", 
                        default=False, action='store_true')

    parser.add_argument('--debug', dest="debug", 
                        help="use verbose logging to debug.", 
                        default=False, action='store_true')

    parser.add_argument("--outfolder","-o", dest='outfolder', 
                        help="full path to save output, will use temporary folder if not specified", 
                        type=str, default=None)

    parser.add_argument("--format","-f", dest='format', 
                        help="format of images, default is dicom", 
                        default="dicom", choices=['dicom'])

    parser.add_argument('--overwrite', dest="overwrite", 
                        help="overwrite pre-existing files in output directory, if they exist.", 
                        default=False, action='store_true')

    subparsers = parser.add_subparsers(help='action for deid to perform',
                                       title='actions',
                                       description='actions for deid to perform',
                                       dest="command")


    version = subparsers.add_parser("version",
                                    help="print version and exit")
 
    # Checks (checks / tests for various services)
    inspect = subparsers.add_parser("inspect",
                                    help="various checks for PHI and quality")

    inspect.add_argument(nargs="+", dest='folder', 
                         help="input folder or single image. If not provided, test data will be used.", 
                         type=str, default=None)

    inspect.add_argument("--deid", dest='deid', 
                         help="deid file with preferences, if not specified, default used.", 
                         type=str, default=None)

    inspect.add_argument("--pattern", dest='pattern', 
                         help="A pattern to match files in input folder.", 
                         type=str, default=None)

    inspect.add_argument('--save',"-s", dest="save", 
                         help="save result to output tab separated file.", 
                         default=False, action='store_true')

    ids = subparsers.add_parser("identifiers",
                                help="extract and replace identifiers from headers")

    ids.add_argument("--deid", dest='deid', 
                     help="deid file with preferences, if not specified, default used.", 
                     type=str, default=None)

    # A path to an ids file, required if user wants to put (without get)
    ids.add_argument("--ids", dest='ids', 
                     help="Path to a json file with identifiers, required for PUT if you don't do get (via all)", 
                     type=str, default=None)

    ids.add_argument("--input", dest='input', 
                     help="Input folder or single image to perform action on.", 
                     type=str, default=None)

    # Action
    ids.add_argument("--action","-a", dest='action', 
                     help="specify to get, put (replace), or all. Default will get identifiers.", 
                     default=None, choices=['get','put','all','inspect'],required=True)

    return parser


def get_subparsers(parser):
    '''get_subparser will get a dictionary of subparsers, to help with printing help
    '''
    actions = [action for action in parser._actions 
               if isinstance(action, argparse._SubParsersAction)]

    subparsers = dict()
    for action in actions:
        # get all subparsers and print help
        for choice, subparser in action.choices.items():
            subparsers[choice] = subparser

    return subparsers



def main():

    parser = get_parser()
    subparsers = get_subparsers(parser)

    try:
        args = parser.parse_args()
    except:
        sys.exit(0)

    if args.command == "version":
        print(__version__)
        sys.exit(0)

    # if environment logging variable not set, make silent
    os.environ['MESSAGELEVEL'] = "INFO"
    if args.quiet is True:
        os.environ['MESSAGELEVEL'] = "QUIET"
        
    # Initialize the message bot, with level above
    from deid.logger import bot

    if args.command == "identifiers":
        from .identifiers import main
    elif args.command == "inspect":
        from .inspect import main
    else:
        parser.print_help()
        sys.exit(1)

    # Run main for selection
    return main(args,parser)


if __name__ == '__main__':
    main()



'''

Copyright (c) 2017-2018 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

from pydicom.sequence import Sequence
from pydicom.dataset import RawDataElement

from deid.logger import bot
from pydicom import read_file
import os


def extract_sequence(sequence,prefix=None):
    '''return a pydicom.sequence.Sequence recursively
       as a list of dictionary items
    '''
    items = []
    for item in sequence:
        for key,val in item.items():
            if not isinstance(val,RawDataElement):
                header = val.keyword
                if prefix is not None:
                    header = "%s__%s" %(prefix,header)  
                value = val.value
                if isinstance(value,bytes):
                    value = value.decode('utf-8')
                if isinstance (value,Sequence):
                    items += extract_sequence(value,prefix=header)
                    continue
                entry = {"key": header, "value": value}
                items.append(entry)
    return items



def expand_field_expression(field,dicom,contenders=None):
    '''Get a list of fields based on an expression. If 
       no expression found, return single field.
    '''
    fields = field.split(':')
    if len(fields) == 1:
        return fields
    expander,expression = fields
    fields = []
    if contenders is None:
        contenders = dicom.dir()
    if expander.lower() == "endswith":
        fields = [x for x in contenders if x.endswith(expression)]
    elif expander.lower() == "startswith":
        fields = [x for x in contenders if x.startswith(expression)]
    return fields



def get_fields(dicom, skip=None, expand_sequences=True):
    '''get fields is a simple function to extract a dictionary of fields
       (non empty) from a dicom file.
    '''    
    if skip is None:
        skip = []
    if not isinstance(skip,list):
        skip = [skip]
    fields = dict()
    contenders = dicom.dir()
    for contender in contenders:
        if contender in skip:
            continue

        try:
            value = dicom.get(contender)
            # Adding expanded sequences
            if isinstance(value,Sequence) and expand_sequences is True:
                sequence_fields = extract_sequence(value,prefix=contender)
                for sf in sequence_fields:
                    fields[sf['key']] = sf['value']
            else:
                if value not in [None,""]:
                    if isinstance(value,bytes):
                        value = value.decode('utf-8')
                    fields[contender] = str(value)
        except:
            pass # need to look into this bug
    return fields



def get_fields_byVR(dicom,exclude_fields=None):
    '''filter a dicom's fields based on a list of value
       representations (VR). If exclude_fields is not defined,
       defaults to "US" and "SS"
    '''

    if exclude_fields is None:
        exclude_fields = ['US','SS']

    if not isinstance(exclude_fields,list):
        exclude_fields = [exclude_fields]

    fields = []
    for field in dicom.dir():
        if dicom.data_element(field) is not None:
            if "VR" in dicom.data_element(field).__dict__:
                if dicom.data_element(field) not in exclude_fields:
                    fields.append(field)
    return fields



from .header import (
    get_identifiers,
    get_shared_identifiers,
    replace_identifiers,
    remove_private_identifiers
)

from .utils import (
    get_files
)

from .fields import  extract_sequence

from .pixels import (
    has_burned_pixels,
    DicomCleaner
)



'''

logger/spinner.py: Simple spinner for logger

Copyright (c) 2017-2018 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import os
import sys

import sys
import time
import threading
from random import choice

class Spinner:
    spinning = False
    delay = 0.1

    @staticmethod
    def spinning_cursor():
        while 1: 
            for cursor in '|/-\\': yield cursor

    @staticmethod
    def balloons_cursor():
        while 1: 
            for cursor in '. o O @ *': yield cursor

    @staticmethod
    def changing_arrows():
        while 1: 
            for cursor in '<^>v': yield cursor

    def select_generator(self, generator):
        if generator == None:
            generator = choice(['cursor',
                                'arrow',
                                'balloons'])

        return generator

    def __init__(self, delay=None, generator=None):
        generator = self.select_generator(generator)

        if generator == 'cursor':
            self.spinner_generator = self.spinning_cursor()
        elif generator == 'arrow':
            self.spinner_generator = self.changing_arrows()
        elif generator == 'balloons':
            self.spinner_generator = self.balloons_cursor()
            if delay is None: delay = 0.2
        else:
            self.spinner_generator = self.spinning_cursor()

        if delay and float(delay):
            self.delay = delay

    def run(self):
        while self.spinning:
            sys.stdout.write(next(self.spinner_generator))
            sys.stdout.flush()
            time.sleep(self.delay)
            sys.stdout.write('\b')
            sys.stdout.flush()

    def start(self):
        self.spinning = True
        threading.Thread(target=self.run).start()

    def stop(self):
        self.spinning = False
        time.sleep(self.delay)



'''

header.py: functions to extract identifiers from dicom headers

Copyright (c) 2017-2018 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''


from deid.logger import bot
from deid.utils import read_json

from .tags import (
    remove_tag,
    remove_sequences
)

from deid.dicom.tags import get_private
from deid.config import DeidRecipe

from pydicom import read_file
from pydicom.errors import InvalidDicomError
import dateutil.parser
import tempfile

from .utils import save_dicom
from .actions import perform_action
from pydicom.dataset import Dataset

from .fields import (
    get_fields,
    get_fields_byVR
)

import os

here = os.path.dirname(os.path.abspath(__file__))


################################################################################
# MAIN GET FUNCTIONS
################################################################################

def get_shared_identifiers(dicom_files,
                           force=True,
                           config=None,
                           aggregate=None,
                           expand_sequences=True):
    '''

    extract shared identifiers across a set of dicom files, intended for
    cases when a set of images (dicom) are being compressed into one file
    and the file (still) should have some searchable metadata. By default,
    we remove fields that differ between files. To aggregate unique, define
    a list of aggregate fields (aggregate).

    '''

    bot.debug('Extracting shared identifiers for %s dicom' %(len(dicom_files)))

    if aggregate is None:
        aggregate = []

    if config is None:
        config = "%s/config.json" %(here)

    if not os.path.exists(config):
        bot.error("Cannot find config %s, exiting" %(config))
    config = read_json(config, ordered_dict=True)['get']

    if not isinstance(dicom_files,list):
        dicom_files = [dicom_files]
    ids = dict() # identifiers

    # We will skip PixelData
    skip = config['skip']
    for dicom_file in dicom_files:
        dicom = read_file(dicom_file,force=True)
        fields = get_fields(dicom,
                            skip=skip,
                            expand_sequences=expand_sequences)
        for key,val in fields.items():

            # If it's there, only keep if the same
            if key in ids:

                # Items to aggregate are appended, not removed
                if key in aggregate:
                    if val not in ids[key]:
                        ids[key].append(val)
                else:

                    # Keep only if equal between
                    if ids[key] == val:
                        continue
                    else:
                        del ids[key]
                        skip.append(key)
            else:
                if key in aggregate:
                    val = [val]
                ids[key] = val

    # For any aggregates that are one item, unwrap again
    for field in aggregate:
        if field in ids:
            if len(ids[field])==1:
                ids[field] = ids[field][0]

    return ids



def get_identifiers(dicom_files,
                    force=True,
                    config=None,
                    expand_sequences=True,
                    skip_fields=None):
    ''' extract all identifiers from a dicom image.
        This function returns a lookup by file name

        Parameters
        ==========
        dicom_files: the dicom file(s) to extract from
        force: force reading the file (default True)
        config: if None, uses default in provided module folder
        expand_sequences: if True, expand sequences. otherwise, skips
        skip_fields: if not None, added fields to skip

    '''
    bot.debug('Extracting identifiers for %s dicom' %(len(dicom_files)))

    if config is None:
        config = "%s/config.json" %(here)

    if not os.path.exists(config):
        bot.error("Cannot find config %s, exiting" %(config))
    config = read_json(config, ordered_dict=True)['get']

    if not isinstance(dicom_files,list):
        dicom_files = [dicom_files]

    ids = dict() # identifiers

    # We will skip PixelData
    skip = config['skip']
    if skip_fields is not None:
        if not isinstance(skip_fields,list):
            skip_fields = [skip_fields]
        skip = skip + skip_fields
 
    for dicom_file in dicom_files:
        dicom = read_file(dicom_file,force=True)

        if dicom_file not in ids:
            ids[dicom_file] = dict()

        ids[dicom_file] = get_fields(dicom,
                                  skip=skip,
                                  expand_sequences=expand_sequences)
    return ids


def remove_private_identifiers(dicom_files,
                               save=True,
                               overwrite=False,
                               output_folder=None,
                               force=True):

    '''remove_private_identifiers is a wrapper for the 
    simple call to dicom.remove_private_tags, it simply
    reads in the files for the user and saves accordingly
    '''
    updated_files = []
    if not isinstance(dicom_files,list):
        dicom_files = [dicom_files]

    for dicom_file in dicom_files:
        dicom = read_file(dicom_file,force=force)
        dicom.remove_private_tags()
        dicom_name = os.path.basename(dicom_file)
        bot.debug("Removed private identifiers for %s" %dicom_file)

        if save:
            dicom = save_dicom(dicom=dicom,
                               dicom_file=dicom_file,
                               output_folder=output_folder,
                               overwrite=overwrite)

        updated_files.append(dicom)
    return updated_files



def _prepare_replace_config(dicom_files, deid=None, config=None):
    ''' replace identifiers will replace dicom_files with data from ids based
        on a combination of a config (default is remove all) and a user deid

        Parameters
        ==========
        dicom_files: the dicom file(s) to extract from
        force: force reading the file (default True)
        save: if True, save to file. Otherwise, return dicom objects
        config: if None, uses default in provided module folder
        overwrite: if False, save updated files to temporary directory
    
    '''

    if config is None:
        config = "%s/config.json" %(here)
    if not os.path.exists(config):
        bot.error("Cannot find config %s, exiting" %(config))
    
    if not isinstance(deid, DeidRecipe):
        deid = DeidRecipe(deid)
    config = read_json(config, ordered_dict=True)

    if not isinstance(dicom_files,list):
        dicom_files = [dicom_files]

    return dicom_files, deid, config


def replace_identifiers(dicom_files,
                        ids,
                        deid=None,
                        save=True,
                        overwrite=False,
                        output_folder=None,
                        force=True,
                        config=None,
                        strip_sequences=True,
                        remove_private=True):

    '''replace identifiers using pydicom, can be slow when writing
    and saving new files'''

    dicom_files, recipe, config = _prepare_replace_config(dicom_files, 
                                                          deid=deid,
                                                          config=config)

    # Parse through dicom files, update headers, and save
    updated_files = []
    for d in range(len(dicom_files)):
        dicom_file = dicom_files[d]
        dicom = read_file(dicom_file,force=force)
        dicom_name = os.path.basename(dicom_file)
        fields = dicom.dir()

        # Remove sequences first, maintained in DataStore
        if strip_sequences is True:
            dicom = remove_sequences(dicom)

        if recipe.deid is not None:
            if dicom_file in ids:
                for action in deid.get_actions():
                    dicom = perform_action(dicom=dicom,
                                           item=ids[dicom_file],
                                           action=action)
            else:
                bot.warning("%s is not in identifiers." %dicom_name)
                continue

        # Next perform actions in default config, only if not done
        for action in config['put']['actions']:
            if action['field'] in fields:
                 dicom = perform_action(dicom=dicom,
                                        action=action)
        if remove_private is True:
            try:
                dicom.remove_private_tags()
            except:
                bot.error('''Private tags for %s could not be completely removed, usually
                             this is due to invalid data type. Removing others.''' % dicom_name)
                private_tags = get_private(dicom)
                for ptag in private_tags:
                    del dicom[ptag.tag]
                continue
        else:
            bot.warning("Private tags were not removed!")

        ds = Dataset()
        for field in dicom.dir():
            try:
                ds.add(dicom.data_element(field))
            except:
                pass

        # Copy original data attributes
        attributes = ['is_little_endian',
                      'is_implicit_VR',
                      'is_decompressed',
                      'read_encoding',
                      'read_implicit_vr',
                      'read_little_endian',
                      '_parent_encoding']

        # We aren't including preamble, we will reset to be empty 128 bytes
        ds.preamble = b"\0" * 128

        for attribute in attributes:
            if hasattr(dicom, attribute):
                ds.__setattr__(attribute, dicom.__getattribute__(attribute))

        # Original meta data                     # or default empty dataset
        file_metas = getattr(dicom, 'file_meta', Dataset())
        
        # Media Storage SOP Instance UID can be identifying
        if hasattr(file_metas, 'MediaStorageSOPInstanceUID'):
            file_metas.MediaStorageSOPInstanceUID = ''

        # Save meta data
        ds.file_meta = file_metas
       
       # Save to file?
        if save is True:
            ds = save_dicom(dicom=ds,
                            dicom_file=dicom_file,
                            output_folder=output_folder,
                            overwrite=overwrite)
        updated_files.append(ds)

    return updated_files



#!/usr/bin/env python3

'''

The MIT License (MIT)

Copyright (c) 2017-2018 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

from deid.logger import bot
from deid.dicom import get_files
from deid.data import get_dataset  
from deid.config import load_deid, get_deid
from deid.dicom import has_burned_pixels

from glob import glob
import datetime
import tempfile
import argparse
import shutil
import sys
import os


def main(args,parser):
    '''inspect currently serves to inspect the header fields of a set
    of dicom files against a standard, and flag images that don't
    pass the different levels of criteria
    '''

    # Global output folder
    #output_folder = args.outfolder
    #if output_folder is None:
    #    output_folder = tempfile.mkdtemp()

    # If a deid is given, check against format
    deid = args.deid
    if deid is not None:
        params = load_deid(deid)
        if params['format'] != args.format:
            bot.error("Format in deid (%s) doesn't match choice here (%s) exiting." %(params['format'],
                                                                                      args.format))
    # Get list of dicom files
    base = args.folder
    if base is None:
        bot.info("No input folder specified, will use demo dicom-cookies.")
        base = get_dataset('dicom-cookies')

    dicom_files = list(get_files(base, pattern=args.pattern)) # todo : consider using generator functionality
    result = has_burned_pixels(dicom_files, deid=deid)

    print('\nSUMMARY ================================\n')
    if len(result['clean']) > 0:
        bot.custom(prefix='CLEAN',
                   message="%s files" %len(result['clean']),
                   color="CYAN")

    if len(result['flagged']) > 0:
        for group,files in result['flagged'].items():
            bot.flag("%s %s files" %(group, len(files)))

    if args.save is True:
        folders = '-'.join([os.path.basename(folder) for folder in base])
        outfile = "pixel-flag-results-%s-%s.tsv" %(folders,datetime.datetime.now().strftime('%y-%m-%d'))
        with open(outfile,'w') as filey:
            filey.writelines('dicom_file\tpixels_flagged\tflag_list\treason\n')
            for clean in result['clean']:
                filey.writelines('%s\tCLEAN\t\t\n' %clean)
            for flagged,details in result['flagged'].items():
                if details['flagged'] is True:
                    for result in details['results']:
                        group = result['group']
                        reason = result['reason']
                        filey.writelines('%s\tFLAGGED\t%s\t%s\n' %(flagged,group,reason))

            print('Result written to %s' %outfile)
            
 
if __name__ == '__main__':
    main()



'''

Copyright (c) 2018 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

from deid.logger import bot
import dateutil.parser
from datetime import timedelta
import os
import re
import sys


def parse_value(item, value, field=None):
    '''parse_value will parse the value field of an action,
    either returning: 
        1. the string (string or from function)
        2. a variable looked up (var:FieldName)
    '''
    # If item is passed as None
    if item == None:
        item = dict()

    # Does the user want a custom value?
    if re.search('[:]', value):
        value_type, value_option = value.split(':') 
        if value_type.lower() == "var": 

            # If selected variable not provided, skip
            if value_option not in item:
                return None
            return item[value_option]

        # The user is providing a specific function
        elif value_type.lower() == "func": 

            if value_option not in item:
                bot.warning('%s not found in item lookup %s' %(value_option))
                return None
            return item[value_option](item, value, field)

        bot.warning('%s is not a valid value type, skipping.' %(value_type))
        return None
    return value



def get_func(function_name):
    '''get_func will return a function that is defined from a string.
       the function is assumed to be in this file

       Parameters
       ==========
       return a function from globals based on a name string

    '''
    env = globals()
    if function_name in env:
        return env[function_name]
    return None


# Timestamps

def get_timestamp(item_date, item_time=None, jitter_days=None, format=None):
    '''get_timestamp will return (default) a UTC timestamp 
    with some date and (optionall) time. A different format can be 
    provided to change default behavior. eg: "%Y%m%d"
    '''
    if format is None:
        format = "%Y-%m-%dT%H:%M:%SZ"

    if item_date in ['', None]:
        bot.warning("No date in header, cannot create timestamp.")
        return None

    if item_time is None:
        item_time = ""

    timestamp = dateutil.parser.parse("%s%s" %(item_date,item_time))
    if jitter_days is not None:
        jitter_days = int(float(jitter_days))
        timestamp = timestamp + timedelta(days=jitter_days)

    return timestamp.strftime(format)
